/**
 * Title:        mike.bid.com
 * Description:
 * Copyright:    Copyright (c) 2002
 * Company:
 * @author Charles E. Hughes
 * @version 1.0
 */

// A bid server

import java.util.Vector;
import java.util.Enumeration;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.RMISecurityManager;
import java.rmi.NotBoundException;
import java.rmi.Naming;
import java.rmi.server.UnicastRemoteObject;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.net.MalformedURLException;


public class BidServerImpl extends UnicastRemoteObject implements BidServer {

    Vector clients = new Vector();
    Vector bidList = new Vector();
    int uidCounter=0;

    public BidServerImpl(String port, String service) throws RemoteException {
      if (System.getSecurityManager() == null) {
        System.setSecurityManager(new RMISecurityManager());
      }
      try {
        String host = InetAddress.getLocalHost().getHostName();
        String url = "rmi://" + host + ":" + port + "/" + service;
        System.err.println("url = " + url);
        Naming.rebind(url, this);
        System.out.println("Server bound to: " + url);
      } catch (UnknownHostException ex) {
        System.err.println("Couldn't get local host name");
        System.exit(1);
      } catch (RemoteException ex) {
        System.err.println("Couldn't contact rmiregistry.");
        System.err.println("Are you sure you're running rmiregistry?");
        ex.printStackTrace();
        System.exit(1);
      } catch (MalformedURLException ex) {
        // Shouldn't happen
        System.exit(1);
      }
    }

    public int register(BidClient client) throws RemoteException {
      // return unique id (-1 if fail)
      System.err.println("Registered: " + client.whoAreYou());
      for (Enumeration e = clients.elements() ; e.hasMoreElements() ;)
        client.friend((BidClient)e.nextElement());
      clients.add(client);
      client.updateBidList(bidList);
      return (uidCounter++);
    }

    public void unregister(int id) throws RemoteException {
      // removes this client from server's registry -- must remove its sale items and update bid list
      Bid bidNode;
      BidClient clientNode;
      int i=0;
      // delete bids from list with that uid
      while (i<bidList.size()) {
        bidNode=(Bid)bidList.elementAt(i);
        if (bidNode.getOwner()==id) {
          bidList.remove(i);
        } else {
          i++;
        }
      }

      // delete client from list
      i=0;
      while (i<clients.size()) {
        clientNode=(BidClient)clients.elementAt(i);
        if (clientNode.getUID()==id) {
          System.err.println("Unregistered: " + clientNode.whoAreYou());
          clients.remove(i);
        } else {
          i++;
        }
      }

    // update clients
    updateClients();
    }

    public boolean offer(int id, String item, int price) throws RemoteException {
      // return true if accepted (max 3 items per client)
      System.out.println("received an offer from id="+id+" on "+item+" for $"+price);
      int i;
      int count=0;
      Bid bidNode;
      Bid dup=null;

      // search through bidList to see if they have < 3 or find a duplicate
      for (i=0; i<bidList.size(); i++) {
        bidNode=(Bid)bidList.elementAt(i);
        if (bidNode.getOwner()==id) count++;
        if (bidNode.getName().equals(item)) dup=bidNode;
      }

      if (count>=3) return(false);
      if (dup!=null) return(false);

      // create bid
      bidNode=new BidImpl(id,item,price);
      // add bid
      bidList.add(bidNode);
      // update clients
      updateClients();
      return(true);
    }

    public boolean bid(int id, String item, int price) throws RemoteException {
      // return true if accepted (> current bid && not sold)
      // search list for that item
      System.out.println("received a bid from id="+id+" on "+item+" for $"+price);
      int i=0;
      Bid bidNode=null;
      Bid targetBid=null;
      int targetPos=0;
      for (i=0; i<bidList.size(); i++) {
        bidNode=(Bid)bidList.elementAt(i);
        if (bidNode.getName().equals(item))  {
          targetBid=bidNode;
          targetPos=i;
        }
      }

      if (targetBid==null) return(false);
      if (targetBid.getPrice()>=price) return(false);
      if (targetBid.isSold()) return(false);

      // everything's okay lets update the bid
      targetBid.setBid(id,price);
      bidList.setElementAt(targetBid,targetPos);
      updateClients();
      return(true);
    }

    public boolean accept(int id, String item) throws RemoteException {
      // reurns true if there is a bid to accept and you are seller
      // search list for that item
      System.out.println("received an accept from id="+id+" on "+item);
      int i=0;
      Bid bidNode=null;
      Bid targetBid=null;
      BidClient buyer;
      BidClient seller;

      for (i=0; i<bidList.size(); i++) {
        bidNode=(Bid)bidList.elementAt(i);
        if (bidNode.getName().equals(item)) targetBid=bidNode;
      }
      if (targetBid==null) return(false);
      if (targetBid.getOwner()!=id) return(false);

      // now, send both clients a message to allow them to conduct sale
      seller=getClientByID(id);
      buyer=getClientByID(targetBid.getHighBidder());
      if (buyer==null) return(false);  // if no one bid on item

      targetBid.justSold();   // everything's okay update bid.

      seller.sale(buyer,targetBid.getName(),targetBid.getPrice());
      buyer.purchase(seller,targetBid.getName(), targetBid.getPrice());
      bidList.remove(targetBid);
      updateClients();
      return(true);
    }

    private void updateClients() {
      // sends the bidList to all the clients
      int i;
      BidClient clientNode;

      for (i=0; i<clients.size(); i++) {
        clientNode=(BidClient)clients.elementAt(i);
        try {
          clientNode.updateBidList(bidList);
        } catch (RemoteException e) {
          System.out.println("problem updating a client");
          e.printStackTrace();
        }
      }
    }

    private BidClient getClientByID(int id) {
      // finds and returns a client by their ID number
      int i;
      BidClient clientNode;
      int cl=0;
      BidClient target=null;

      for (i=0; i<clients.size(); i++) {
        clientNode=(BidClient)clients.elementAt(i);
        try {
          cl=clientNode.getUID();
        } catch (RemoteException e) {
          System.out.println("problem talking to a client");
          e.printStackTrace();
        }
        if (cl==id) target=clientNode;
      }
      return(target);
    }

    private void printBidList() {
      int i;
      Bid bidNode;

      for (i=0; i<bidList.size(); i++) {
        bidNode=(Bid)bidList.elementAt(i);
        System.out.println("["+bidNode.getName()+","+bidNode.getPrice()+"]");
      }
    }

    public static void main(String[] args) {
      String service;
      String port;
      if (args.length > 0) service = args[0];
      else service = "BidServer";
      if (args.length > 1) port = args[1];
      else port = "1099";
      try {
        BidServerImpl server = new BidServerImpl(port,service);
      } catch (RemoteException ex) {
        System.err.println("Trouble creating server: " + ex.getMessage());
        ex.printStackTrace();
      }
   }
}
